import threading

from fastapi import FastAPI
from loguru import logger
from starlette.middleware.cors import CORSMiddleware
from uvicorn import Config, Server


class WebServer:
    def __init__(self, port: int):
        self._app = self._setup_app()
        self._server_handle = None
        self._uvicorn = None
        self._port = port

    def _setup_app(self) -> FastAPI:
        app = FastAPI(title="Plugin Webserver", redoc_url=None, swagger_ui_oauth2_redirect_url=None)

        # TODO setup allow_origin_regex for *.memri.io and localhost
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_methods=["*"],
            allow_headers=["*"],
        )

        return app

    @property
    def app(self) -> FastAPI:
        return self._app

    def run(self):
        """Starts the webserver, if not done already.
        Call returns immediately, server itself is offloaded to
        the thread.
        """

        # Bare application has two endpoints registered:
        # /openapi.json and /docs
        # Rest is registered by the plugin owner
        if not self.is_running():
            config = Config(app=self.app, host="0.0.0.0", port=self._port, workers=1)
            self._uvicorn = Server(config=config)

            self._server_handle = threading.Thread(target=self._uvicorn.run, daemon=False)
            self._server_handle.start()

    def is_running(self):
        return self._server_handle is not None

    def shutdown(self):
        """Shuts down the uvicorn server, frees the thread"""
        logger.info("Shutting down the webserver..")
        if self._uvicorn:
            self._uvicorn.should_exit = True
            # self._uvicorn.force_exit = True
            logger.info("Joining the thread...")

            if self._server_handle:
                self._server_handle.join()
                self._server_handle = None

        logger.info("Webserver shut down")



import os

from fastapi import APIRouter, Request
from fastapi.responses import HTMLResponse, JSONResponse
from fastapi.templating import Jinja2Templates

from .helpers import DictProxy, get_root_path

qr_code_dict = DictProxy()

# Jinja2Templates class resolves relative paths to current working directory,
# which is not pymemri/webserver, but path to the plugin that uses it.
# Flask is clever enough to be aware that application may be a part of
# module or package, and changes relative paths to absolute, the first parameter
# of Flask app helps with that:
# app = flask.Flask(__name__, template_folder='template')
# Read the Flask code for better understanding.
# In order to have the same behavior for FastApi get_root_path function is used.
# that is actually copy pasted from Flask.
templates = Jinja2Templates(directory=os.path.join(get_root_path(__name__), "template"))

router = APIRouter()

QR_CODE_KEY = "qr_code"
QR_STRING_KEY = "qr_string"
AUTHENTICATED = "authenticated"
REQUIRES2FACTOR = "requires2factor"


@router.get("/qr", response_class=HTMLResponse)
def qr(request: Request):
    """Returns rendered HTML with QR image"""
    global qr_code_dict
    qr_code_data = qr_code_dict.get(QR_CODE_KEY, None)
    done = qr_code_dict.get(AUTHENTICATED, False)
    requires2factor = (qr_code_dict.get(REQUIRES2FACTOR, False),)

    the_path = os.path.join(get_root_path(__name__), "template")

    # TODO - add a template for 2FA
    if done:
        return templates.TemplateResponse("success.html", {"request": request})
    else:
        return templates.TemplateResponse(
            "images.html", {"request": request, "chart_output": qr_code_data}
        )


@router.get("/qr_svg", response_class=JSONResponse)
def qr_svg():
    """Returns the QR svg as json"""
    global qr_code_dict

    content = {
        "qr": qr_code_dict.get(QR_CODE_KEY, None),
        "authenticated": qr_code_dict.get(AUTHENTICATED, False),
        "requires2factor": qr_code_dict.get(REQUIRES2FACTOR, False),
    }

    return JSONResponse(content=content, headers={"Access-Control-Allow-Origin": "*"})


@router.get("/qr_string", response_class=JSONResponse)
def qr_string():
    """Returns the QR svg as json"""
    global qr_code_dict

    content = {
        "qr": qr_code_dict.get(QR_STRING_KEY, None).decode("utf-8"),
        "authenticated": qr_code_dict.get(AUTHENTICATED, False),
        "requires2factor": qr_code_dict.get(REQUIRES2FACTOR, False),
    }

    return JSONResponse(content=content, headers={"Access-Control-Allow-Origin": "*"})



# run states
RUN_IDLE = "idle"
RUN_INITIALIZED = "initialized"
RUN_USER_ACTION_NEEDED = "userActionNeeded"
RUN_USER_ACTION_COMPLETED = "ready"
RUN_STARTED = "started"
RUN_FAILED = "error"
RUN_COMPLETED = "done"
RUN_ABORTED = "aborted"
RUN_AUTHENTICATED = "authenticated"
RUN_DAEMON = "daemon"



import time

from fastscript import Param, call_parse
from loguru import logger

from pymemri.plugin.states import (
    RUN_COMPLETED,
    RUN_FAILED,
    RUN_STARTED,
    RUN_USER_ACTION_COMPLETED,
    RUN_USER_ACTION_NEEDED,
)
from pymemri.pod.client import PodClient

SLEEP_INTERVAL = 1


def input_credentials():
    username = input("Enter username: ")
    password = input("Enter password: ")
    return username, password


@call_parse
def run_password_simulator(
    database_key: Param("database key", str) = None,
    owner_key: Param("owner key", str) = None,
    run_id: Param("owner key", str) = None,
):

    print("FRONTEND PASSWORD SIMULATOR")
    if database_key and owner_key:
        logger.info("pod credentials from from args")
        client = PodClient(database_key=database_key, owner_key=owner_key)
    else:
        client = PodClient.from_local_keys()

    while True:
        time.sleep(SLEEP_INTERVAL)
        pluginRun = client.get(run_id)
        account = pluginRun.account[0]

        if pluginRun.status == RUN_USER_ACTION_NEEDED:
            username, password = input_credentials()
            account.identifier = username
            account.secret = password
            pluginRun.status = "ready"
            client.update_item(pluginRun)
            client.update_item(account)
            logger.info("Done authenticating")
            break

        elif pluginRun.status == RUN_STARTED:
            logger.info("plugin starting...")

        elif pluginRun.status == RUN_USER_ACTION_COMPLETED:
            logger.info("no user action needed.")

        elif pluginRun.status == RUN_FAILED:
            logger.error("error occurred in plugin.")
            break

        elif pluginRun.status == RUN_COMPLETED:
            break

        else:
            logger.warning(f"unknown plugin state {pluginRun.status}.")



# from pymemri.importers.email import EmailImporter
# from pymemri.indexers.geo.geo_indexer import GeoIndexer
# from pymemri.indexers.notelist.notelist_indexer import NotesListIndexer
# from pymemri.indexers.faceclustering.indexer import FaceClusteringIndexer



from datetime import datetime
from typing import List, Optional, Union

from pydantic import PrivateAttr

from pymemri.data.schema import Item, Message, Person, Tweet


class Dog(Item):
    _hasBeenFed: bool = PrivateAttr(False)

    # Properties
    name: Optional[str] = None
    age: Optional[int] = None
    bites: Optional[bool] = None
    weight: Optional[float] = None
    dateOfBirth: Optional[datetime] = None

    # Edges
    owner: List[Person] = []
    referencedBy: List[Union[Message, Tweet]] = []



import os
from pathlib import Path


def get_ci_variables(*varnames):
    """
    Gets ci variables from environment, can be set on gitlab in settings -> ci/cd -> variables
    """
    return {varname: os.environ.get(varname) for varname in varnames}


def get_project_root():
    return Path(__file__).parent.parent



import os

import pytest

from pymemri.gitlab_api import GitlabAPI
from pymemri.pod.client import PodClient
from pymemri.template.formatter import gitlab_slugify


@pytest.mark.parametrize(
    "input,expected",
    [
        ("SAMPLE Project", "sample-project"),
        ("--sample space project--", "sample-space-project"),
        ("s" * 64, "s" * 63),
        ("sample project", "sample-project"),
        ("sample_project", "sample-project"),
        ("01-09-test", "01-09-test"),
    ],
)
def test_gitlab_slugify(input, expected):
    assert gitlab_slugify(input) == expected


@pytest.mark.skipif(
    os.environ.get("CI_JOB_TOKEN") is None,
    reason="CI_JOB_TOKEN environment variable not set",
)
def test_gitlab_project_id():
    client = PodClient()
    gitlab = GitlabAPI(client)
    project_id = gitlab.project_id_from_name("pymemri", "memri")
    assert project_id == 80



import os
import re
import sys
import urllib
from getpass import getpass
from pathlib import Path

import requests
from git import Repo
from loguru import logger

from pymemri.pod.client import PodClient

from .template.formatter import _plugin_from_template, gitlab_slugify

MEMRI_PATH = Path.home() / ".memri"
MEMRI_GITLAB_BASE_URL = "https://gitlab.memri.io"
ACCESS_TOKEN_PATH = Path.home() / ".memri/access_token/access_token.txt"
GITLAB_API_BASE_URL = "https://gitlab.memri.io/api/v4"
DEFAULT_PACKAGE_VERSION = "0.0.1"

TIME_FORMAT_GITLAB = "%Y-%m-%dT%H:%M:%S.%fZ"
PROJET_ID_PATTERN = '(?<=<span class="gl-button-text">Project ID: )[0-9]+(?=</span>)'


class GitlabAPI:
    def __init__(self, client: PodClient = None, request_auth_if_needed: bool = False):
        self.client = client
        self.auth_headers = dict()
        self.auth_params = dict()
        self.auth_initialized = False
        self.request_auth_if_needed = request_auth_if_needed
        self.get_registry_params_headers()

    def init_auth_params(self):
        access_token = self.client.get_oauth2_access_token("gitlab")
        if access_token is None:
            raise RuntimeError(
                "No gitlab access token found in the pod. "
                "Please authenticate with gitlab in the frontend."
            )
        self.auth_params = {"access_token": access_token}

    def get_registry_params_headers(self):
        job_token = os.environ.get("CI_JOB_TOKEN", None)
        if job_token is not None:
            self.auth_initialized = True
            self.auth_headers = {"JOB-TOKEN": job_token}
        elif self.client is not None:
            self.init_auth_params()
            self.auth_initialized = True
        else:
            ACCESS_TOKEN_PATH.parent.mkdir(parents=True, exist_ok=True)
            if ACCESS_TOKEN_PATH.is_file():
                self.auth_initialized = True
                with open(ACCESS_TOKEN_PATH, "r") as f:
                    self.auth_headers = {"PRIVATE-TOKEN": f.read()}
            else:
                if self.request_auth_if_needed:
                    print(
                        """
                    The first time you are uploading a model you need to create an access_token
                    at https://gitlab.memri.io/-/profile/personal_access_tokens?name=Model+Access+token&scopes=api
                    Click at the blue button with 'Create personal access token'"
                    """
                    )
                    access_token = getpass(
                        "Then copy your personal access token from 'Your new personal access token', and paste here: "
                    )
                    with open(ACCESS_TOKEN_PATH, "w") as f:
                        f.write(access_token)
                    self.auth_headers = {"PRIVATE-TOKEN": access_token}
                    self.auth_initialized = True

        # test if we are authenticated
        res = self.get(
            f"{GITLAB_API_BASE_URL}/projects", headers=self.auth_headers, params=self.auth_params
        )
        if res.status_code not in [200, 201]:
            logger.error(res.content)
            raise RuntimeError(res.content)

    def write_file_to_package_registry(
        self,
        project_id,
        file_path,
        package_name,
        version=DEFAULT_PACKAGE_VERSION,
        trigger_pipeline=False,
    ):
        file_path = Path(file_path)
        file_name = file_path.name

        url = f"{GITLAB_API_BASE_URL}/projects/{project_id}/packages/generic/{package_name}/{version}/{file_name}"
        logger.info(f"uploading {file_path}")
        it = upload_in_chunks(file_path)
        res = self.put(
            url=url,
            data=IterableToFileAdapter(it),
            headers=self.auth_headers,
            params=self.auth_params,
        )

        if res.status_code not in [200, 201]:
            logger.error(f"Failed to upload {file_path}: {res.content}")
        else:
            logger.info(f"Succesfully uploaded {file_path}")
            if trigger_pipeline:
                self.trigger_pipeline(project_id)

    def _authenticated_request(self, request_function, *args, **kwargs):
        try:
            return request_function(*args, **kwargs)
        except Exception as e:
            logger.error(e)
            if self.auth_params:
                logger.debug("Maybe access_token is expired, trying again with new auth params")
                self.init_auth_params()
                return request_function(*args, **kwargs)
            else:
                raise e

    def get(self, *args, **kwargs):
        return self._authenticated_request(requests.get, *args, **kwargs)

    def post(self, *args, **kwargs):
        return self._authenticated_request(requests.post, *args, **kwargs)

    def put(self, *args, **kwargs):
        return self._authenticated_request(requests.put, *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._authenticated_request(requests.delete, *args, **kwargs)

    def trigger_pipeline(self, project_id):
        url = f"{GITLAB_API_BASE_URL}/projects/{project_id}/pipeline?ref=main"
        res = self.post(url, headers=self.auth_headers, params=self.auth_params)
        if res.status_code not in [200, 201]:
            logger.error("Failed to trigger pipeline")

    def project_id_from_name(self, project_name, username=None):
        iden = gitlab_slugify(project_name)
        username = username or self.get_current_username()
        uri_encoded_name = urllib.parse.quote_plus(f"{username}/{iden}")
        res = self.get(
            f"{GITLAB_API_BASE_URL}/projects/{uri_encoded_name}",
            headers=self.auth_headers,
            params=self.auth_params,
        )

        if res.status_code not in [200, 201]:
            logger.error(res.content)
            raise RuntimeError(f"Failed to get project id for {project_name}")
        project_id = res.json().get("id", None)
        if project_id:
            return project_id
        else:
            raise ValueError(
                f"No plugin found with name {project_name}, make sure to enter the name as specified in the url of the repo"
            )

    def get_project_id_from_project_path_unsafe(self, project_path):
        try:
            res = self.get(f"{MEMRI_GITLAB_BASE_URL}/{project_path}")
            html = str(res.content)
            match = re.search(PROJET_ID_PATTERN, html)
            return match.group()
        except Exception:
            raise ValueError(f"Could not find project with name {project_path}")

    def download_package_file(
        self,
        filename,
        project_path,
        package_name,
        out_dir=None,
        package_version=DEFAULT_PACKAGE_VERSION,
        download_if_exists=False,
    ):

        project_name = str(project_path).split("/")[-1]
        out_dir = out_dir if out_dir is not None else MEMRI_PATH / "projects" / project_name
        out_dir.mkdir(parents=True, exist_ok=True)

        project_id = self.get_project_id_from_project_path_unsafe(project_path)

        file_path = out_dir / filename

        if file_path.exists() and not download_if_exists:
            logger.warning(
                f"{file_path} already exists, and `download_if_exists`==False, using cached version"
            )
            return file_path

        logger.info(f"downloading {filename} from project {project_path}, package {package_name}")

        res = self.get(
            url=f"{GITLAB_API_BASE_URL}/projects/{project_id}/packages/generic/{package_name}/{package_version}/{filename}"
        )
        res.raise_for_status()
        with open(out_dir / filename, "wb") as f:
            logger.info(f"writing {filename} to {out_dir}")
            f.write(res.content)
        return file_path

    def create_repo(self, repo_name, client=None):
        url = f"{GITLAB_API_BASE_URL}/projects/"
        payload = {"name": repo_name}
        res = self.post(url=url, json=payload, headers=self.auth_headers, params=self.auth_params)

        if res.status_code not in [200, 201]:
            raise ValueError(f"failed to create repo:\n {res.text}")
        logger.info(f"created project {repo_name}")

    def get_current_username(self, client=None):
        url = f"{GITLAB_API_BASE_URL}/user/"
        res = self.get(url=url, headers=self.auth_headers, params=self.auth_params)
        if res.status_code not in [200, 201]:
            raise ValueError(f"Could not find current user {res.content}")
        else:
            username = res.json()["username"]
            return username

    # NEVER EXPORT THIS
    def delete_project(self, path_or_id, client=None):
        url_escape_id = urllib.parse.quote(str(path_or_id), safe="")
        url = f"{GITLAB_API_BASE_URL}/projects/{url_escape_id}"
        res = self.delete(url=url, headers=self.auth_headers, params=self.auth_params)
        if res.status_code not in [200, 201, 202]:
            raise ValueError(f"failed to delete repo:\n {res.text}")
        logger.info(f"deleted project {path_or_id}")

    def commit_file(self, project_name, path_in2out, branch="main", client=None):
        project_id = self.project_id_from_name(project_name)
        #     file_out_path_escaped = urllib.parse.quote(file_out_path, safe='')

        actions = []

        for file_in_path, file_out_path in path_in2out.items():
            content = Path(file_in_path).read_text()
            action_payload = {
                "action": "create",
                "file_path": file_out_path,
                "content": content,
            }
            actions.append(action_payload)

        url = f"{GITLAB_API_BASE_URL}/projects/{project_id}/repository/commits"
        payload = {
            "branch": branch,
            "commit_message": "automated commit",
            "content": content,
            "actions": actions,
        }

        res = self.post(url=url, json=payload, headers=self.auth_headers, params=self.auth_params)
        files_in = list(path_in2out.keys())
        if res.status_code not in [200, 201, 202]:
            raise ValueError(f"failed to make commit with files {files_in}:\n {res.text}")
        logger.info(f"committed files {files_in}")

    def write_files_to_git(self, repo, target_dir, **kwargs):
        path_in2out = dict()
        for p in target_dir.rglob("*"):
            if p.is_file():
                path_in_repo = p.relative_to(target_dir)
                path_in2out[str(p)] = str(path_in_repo)
        self.commit_file(str(repo), path_in2out, **kwargs)

    def create_new_project(self, project_name, user=None, template_url=None):
        tmp_dir = Path("/tmp/test") / project_name
        rm_tree(tmp_dir)
        repo_url = (
            f"{MEMRI_GITLAB_BASE_URL}/{user}/{project_name}"
            if user is not None
            else f"{MEMRI_GITLAB_BASE_URL}/plugins/{project_name}"
        )

        _plugin_from_template(
            template_name="classifier_plugin",
            description="A transformer based sentiment analyis plugin",
            install_requires="transformers,sentencepiece,protobuf,torch==1.10.0",
            target_dir=str(tmp_dir),
            repo_url=repo_url,
            verbose=False,
            user=user,
            template_url=template_url,
        )
        self.write_files_to_git(project_name, tmp_dir)


def find_git_repo():
    path = "."
    for i in range(10):
        try:
            repo = Repo(f"{path + ('.' * i)}/")
        except:
            pass
        else:
            break
    if i == 9:
        raise ValueError(f"could not fine git repo in {os.path.abspath('')}")

    repo_name = repo.remotes.origin.url.split(".git")[0].split("/")[-1]
    return repo_name


class upload_in_chunks(object):
    def __init__(self, filename, chunksize=1 << 14):
        self.filename = filename
        self.chunksize = chunksize
        self.totalsize = os.path.getsize(filename)
        self.readsofar = 0

    def __iter__(self):
        n = 100
        i = 1
        delta = 1 / n

        with open(self.filename, "rb") as file:
            while True:
                data = file.read(self.chunksize)
                if not data:
                    sys.stderr.write("\n")
                    break
                self.readsofar += len(data)
                percent = self.readsofar * 1e2 / self.totalsize
                while (percent / 100) > i * delta:
                    i += 1
                yield data

    def __len__(self):
        return self.totalsize


class IterableToFileAdapter(object):
    def __init__(self, iterable):
        self.iterator = iter(iterable)
        self.length = len(iterable)

    def read(self, size=-1):  # TBD: add buffer for `len(data) > size` case
        return next(self.iterator, b"")

    def __len__(self):
        return self.length


def rm_tree(pth):
    pth = Path(pth)
    for child in pth.glob("*"):
        if child.is_file():
            child.unlink()
        else:
            rm_tree(child)

    try:
        pth.rmdir()
    except FileNotFoundError as e:
        pass



import random
import warnings
from datetime import datetime
from threading import Thread
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

import numpy as np
from loguru import logger
from typing_extensions import Unpack

from pymemri.data.schema.itembase import ItemBase
from pymemri.data.schema.schema import SchemaMeta

from ..data.schema import Account, File, Photo, PluginRun, get_schema_cls
from .api import DEFAULT_POD_ADDRESS, POD_VERSION, PodAPI, PodError
from .db import DB, Priority
from .graphql_utils import GQLQuery
from .utils import DEFAULT_POD_KEY_PATH, read_pod_key


class PodClient:
    # Mapping from python type to schema type
    # TODO move to data.schema once schema is refactored
    TYPE_TO_SCHEMA = {
        bool: "Bool",
        str: "Text",
        int: "Integer",
        float: "Real",
        datetime: "DateTime",
    }

    def __init__(
        self,
        url=DEFAULT_POD_ADDRESS,
        version=POD_VERSION,
        database_key=None,
        owner_key=None,
        auth_json=None,
        verbose=False,
        default_priority=Priority.local,
        create_account=True,
    ):
        self.verbose = verbose
        self.database_key = database_key if database_key is not None else self.generate_random_key()
        self.owner_key = owner_key if owner_key is not None else self.generate_random_key()
        self.api = PodAPI(
            database_key=self.database_key,
            owner_key=self.owner_key,
            url=url,
            version=version,
            auth_json=auth_json,
            verbose=verbose,
        )

        self.default_priority = Priority(default_priority)

        if create_account:
            self.create_account()

        self.api.test_connection()
        self.local_db = DB()
        self.registered_classes = dict()

    @classmethod
    def from_local_keys(cls, path=DEFAULT_POD_KEY_PATH, **kwargs):
        return cls(
            database_key=read_pod_key("database_key"),
            owner_key=read_pod_key("owner_key"),
            **kwargs,
        )

    @staticmethod
    def generate_random_key():
        return "".join([str(random.randint(0, 9)) for i in range(64)])

    def create_account(self):
        try:
            self.api.create_account()
        except PodError as e:
            logger.warning(e)

    def add_to_store(self, item: ItemBase, priority: Priority = None) -> ItemBase:
        item.create_id_if_not_exists()
        priority = priority if priority is not None else self.default_priority
        return self.local_db.merge(item, priority)

    def reset_local_db(self):
        self.local_db = DB()

    def create(self, item):
        self.add_to_store(item)
        try:
            result = self.api.create_item(item.to_json())
            item.id = result
            item.reset_local_sync_state()
            # try:
            #     res= getattr(item, "requires_client_ref", False)
            # except Exception:
            #     pass
            # finally:
            #     res = False

            if hasattr(item, "requires_client_ref") and item.requires_client_ref:
                item._client = self
            return True
        except Exception as e:
            logger.error(e)
            return False

    def create_photo(self, photo, asyncFlag=True):
        file = photo.file[0]

        # create the photo
        items_edges_success = self.bulk_action(
            create_items=[photo, file], create_edges=photo.get_edges("file")
        )
        if not items_edges_success:
            raise ValueError("Could not create file or photo item")

        return self._upload_image(photo.data, asyncFlag=asyncFlag)

    def add_to_schema(self, meta: SchemaMeta, *items: Unpack[Union[Type[ItemBase], ItemBase]]):
        nodes = {}
        edges = {}
        for item in items:
            if not (isinstance(item, ItemBase) or issubclass(item, ItemBase)):
                raise ValueError(f"{item} is not an instance or subclass of Item")
            item_schema = item.pod_schema()
            nodes[item_schema.name] = {"properties": item_schema.properties}

            for (edge_name, edge) in item_schema.edges.items():
                edges.setdefault(edge_name, []).extend(edge)

        self.api.create_schema({"nodes": nodes, "edges": edges, "meta": meta.to_json()})

        for item in items:
            item_cls = type(item) if isinstance(item, ItemBase) else item
            self.registered_classes[item.__name__] = item_cls
        return True

    def _upload_image(self, img, asyncFlag=True, callback=None):
        if isinstance(img, np.ndarray):
            return self.upload_file(img.tobytes(), asyncFlag=asyncFlag, callback=callback)
        elif isinstance(img, bytes):
            return self.upload_file(img, asyncFlag=asyncFlag, callback=callback)
        else:
            raise ValueError(f"Unknown image data type {type(img)}")

    def upload_file(self, file, asyncFlag=True, callback=None):
        if asyncFlag:
            return self.upload_file_async(file, callback=callback)
        try:
            self.api.upload_file(file)
            return True
        except PodError as e:
            # 409 = CONFLICT, file already exists
            if e.status == 409:
                return True
            return False

    def upload_file_async(self, file, callback=None):
        def thread_fn(file, callback):
            result = self.upload_file(file, asyncFlag=False)
            if callback:
                callback(result)

        Thread(target=thread_fn, args=(file, callback)).start()
        return True

    def get_file(self, sha):
        return self.api.get_file(sha)

    def get_photo(self, id, size=640):
        photo = self.get(id)
        self._load_photo_data(photo, size=size)
        return photo

    def _load_photo_data(self, photo, size=None):
        if len(photo.file) > 0 and photo.data is None:
            file = self.get_file(photo.file[0].sha256)
            if file is None:
                logger.error(
                    f"Could not load data of {photo} attached file item does not have data in pod"
                )
                return
            photo.data = file
        else:
            logger.error(f"could not load data of {photo}, no file attached")

    def create_if_external_id_not_exists(self, item):
        if not self.external_id_exists(item):
            self.create(item)

    def external_id_exists(self, item):
        if item.externalId is None:
            return False
        existing = self.search({"externalId": item.externalId})
        return len(existing) > 0

    def create_edges(self, edges):
        return self.bulk_action(create_edges=edges)

    def delete_items(self, items):
        return self.bulk_action(delete_items=items)

    def delete_edge(self, edge):
        edge = self.get_delete_edge_dict(edge)
        try:
            self.api.delete_edge(edge)
            return True
        except PodError as e:
            logger.warning(e)
            return False

    def delete_file(self, id):
        try:
            self.api.delete_file(id)
            return True
        except PodError as e:
            logger.warning(f"Could not delete file {id}: {e}")
            return False

    @staticmethod
    def gather_batch(items, start_idx, start_size=0, max_size=5000000):
        idx = start_idx
        total_size = start_size
        batch_items = []
        for i, x in enumerate(items):
            if i < idx:
                continue
            elif len(str(x)) > max_size:
                idx = i + 1
                logger.error("Could not add item: Item exceeds max item size")
            elif total_size + len(str(x)) < max_size:
                batch_items.append(x)
                total_size += len(str(x))
                idx = i + 1
            else:
                break
        return batch_items, idx, total_size

    def bulk_action(
        self,
        create_items=None,
        update_items=None,
        create_edges=None,
        delete_items=None,
        partial_update=True,
        priority=None,
    ):
        priority = Priority(priority) if priority else None

        all_items = []
        if create_items:
            all_items += create_items
        if update_items:
            all_items += update_items

        # we need to add to local_db to not lose reference.
        if create_items is not None:
            for c in create_items:
                try:
                    res = getattr(c, "requires_client_ref", False)
                except Exception:
                    pass
                finally:
                    res = False
                if res:
                    c._client = self
                if not self.local_db.contains(c):
                    self.add_to_store(c, priority=priority)

        create_items = [item.to_json() for item in create_items] if create_items is not None else []
        update_items = (
            [self.get_update_dict(i, partial_update=partial_update) for i in update_items]
            if update_items is not None
            else []
        )
        create_edges = (
            [self.get_create_edge_dict(i) for i in create_edges] if create_edges is not None else []
        )

        # Note: skip delete_items without id, as items that are not in pod cannot be deleted
        delete_items = (
            [item.id for item in delete_items if item.id is not None]
            if delete_items is not None
            else []
        )

        n_total = len(create_items + update_items + create_edges + delete_items)
        n = 0

        i_ci, i_ui, i_ce, i_di = 0, 0, 0, 0
        while not (
            i_ci == len(create_items)
            and i_ui == len(update_items)
            and i_ce == len(create_edges)
            and i_di == len(delete_items)
        ):
            batch_size = 0
            create_items_batch, i_ci, batch_size = self.gather_batch(
                create_items, i_ci, start_size=batch_size
            )
            update_items_batch, i_ui, batch_size = self.gather_batch(
                update_items, i_ui, start_size=batch_size
            )
            delete_items_batch, i_di, batch_size = self.gather_batch(
                delete_items, i_di, start_size=batch_size
            )
            if i_ci == len(create_items):
                create_edges_batch, i_ce, batch_size = self.gather_batch(
                    create_edges, i_ce, start_size=batch_size
                )
            else:
                create_edges_batch = []
            n_batch = len(
                create_items_batch + update_items_batch + create_edges_batch + delete_items_batch
            )
            n += n_batch
            logger.info(f"BULK: Writing {n}/{n_total} items/edges")

            try:
                result = self.api.bulk(
                    create_items_batch,
                    update_items_batch,
                    create_edges_batch,
                    delete_items_batch,
                )
            except PodError as e:
                logger.error(f"could not complete bulk action {e}, aborting")
                return False
        logger.info(f"Completed Bulk action, written {n} items/edges")

        for item in all_items:
            item.reset_local_sync_state()
        return True

    def get_create_edge_dict(self, edge):
        return {
            "_source": edge.source.id,
            "_target": edge.target.id,
            "_name": edge.name,
        }

    def get_delete_edge_dict(self, edge):
        return {"_source": edge.source.id, "_target": edge.target.id, "_name": edge.name}

    def create_edge(self, edge):
        edge_dict = self.get_create_edge_dict(edge)

        try:
            self.api.create_edge(edge_dict)
            return True
        except PodError as e:
            logger.error(e)
            return False

    def get(self, id, expanded=True, include_deleted=False):
        if not expanded:
            res = self._get_item_with_properties(id)
        else:
            res = self._get_item_expanded(id)
        if res is None:
            raise ValueError(f"Item with id {id} does not exist")
        elif res.deleted and not include_deleted:
            logger.info(f"Item with id {id} has been deleted")
            return
        return res

    def filter_deleted(self, items):
        return [i for i in items if not i.deleted == True]

    def _get_item_expanded(self, id):
        item = self._get_item_with_properties(id)
        edges = self.get_edges(id)
        if item is None or edges is None:
            return None
        for e in edges:
            if e["name"] in item.edges:
                item.add_edge(e["name"], e["item"])
            else:
                logger.debug(f"Could not add edge {e['name']}: Edge is not defined on Item.")
        return item

    def get_edges(self, id):
        try:
            result = self.api.get_edges(id)
            for edge_name in result:
                edge_item = self.item_from_json(edge_name["item"])
                edge_item.reset_local_sync_state()
                edge_name["item"] = edge_item
            return result
        except PodError as e:
            logger.error(e)
            return

    def _get_item_with_properties(self, id):
        try:
            result = self.api.get_item(str(id))
            if not len(result):
                return
            item = self.item_from_json(result[0])
            item.reset_local_sync_state()
            return item
        except PodError as e:
            logger.error(e)
            return

    def get_update_dict(self, item: ItemBase, partial_update: bool = True):
        properties = item.property_dict()
        properties.pop("deleted", None)
        if partial_update:
            properties = {
                k: v for k, v in properties.items() if k == "id" or k in item._updated_properties
            }
        return properties

    def update_item(self, item, partial_update=True):
        data = self.get_update_dict(item, partial_update=partial_update)
        try:
            self.api.update_item(data)
            item.reset_local_sync_state()
            return True
        except PodError as e:
            logger.error(e)
            return False

    def exists(self, id):
        try:
            result = self.api.get_item(str(id))
            if isinstance(result, list) and len(result) > 0:
                return True
            return False
        except PodError as e:
            logger.error(e)
            return False

    def search_paginate(
        self,
        fields_data,
        limit=50,
        include_edges=True,
        add_to_local_db: bool = True,
        priority=None,
    ):
        priority = Priority(priority) if priority else None

        if "ids" in fields_data:
            raise NotImplementedError(
                "Searching by multiple IDs is not implemented for paginated search."
            )

        extra_fields = {"[[edges]]": {}} if include_edges else {}
        query = {**fields_data, **extra_fields}

        try:
            for page in self.api.search_paginate(query, limit):
                result = [
                    self._item_from_search(item, add_to_local_db=add_to_local_db, priority=priority)
                    for item in page
                ]
                yield self.filter_deleted(result)
        except PodError as e:
            logger.error(e)

    T = TypeVar("T", bound=ItemBase)

    def search_typed(
        self,
        item_type: Type[T],
        fields_data: Optional[Dict[str, Any]] = None,
        *args,
        **kwargs,
    ) -> List[T]:
        fields_data = fields_data or {}

        item_type_name = item_type.__qualname__
        if "." in item_type_name:
            item_type_name = item_type.__qualname__.split(".")[-1]
            logger.warning(
                f"Item type {item_type_name} is not a top-level type."
                f"Searching for {item_type_name} instead"
            )
        return self.search({**fields_data, "type": item_type_name}, *args, **kwargs)

    def search(
        self,
        fields_data,
        include_edges: bool = True,
        add_to_local_db: bool = True,
        priority=None,
    ):
        priority = Priority(priority) if priority else None

        extra_fields = {"[[edges]]": {}} if include_edges else {}
        query = {**fields_data, **extra_fields}
        result = []

        # Special key "ids" for searching a list of ids.
        # Requires /bulk search instead of /search.
        if "ids" in query:
            ids = query.pop("ids")
            bulk_query = [{"id": uid, **query} for uid in ids]
            try:
                result = self.api.bulk(search=bulk_query)["search"]
            except PodError as e:
                logger.error(e)

            result = [item for sublist in result for item in sublist]
        else:
            try:
                result = self.api.search(query)
            except PodError as e:
                logger.error(e)

        result = [
            self._item_from_search(item, add_to_local_db=add_to_local_db, priority=priority)
            for item in result
        ]
        return self.filter_deleted(result)

    def _item_from_search(self, item_json: dict, add_to_local_db: bool = True, priority=None):
        # search returns different fields w.r.t. edges compared to `get` api,
        # different method to keep `self.get` clean.
        item = self.item_from_json(item_json, add_to_local_db=add_to_local_db, priority=priority)
        item.reset_local_sync_state()

        for edge_json in item_json.get("[[edges]]", []):
            edge_name = edge_json["_edge"]
            try:
                edge_item = self.item_from_json(
                    edge_json["_item"],
                    add_to_local_db=add_to_local_db,
                    priority=priority,
                )
                edge_item.reset_local_sync_state()
                item.add_edge(edge_name, edge_item)
            except Exception as e:
                logger.error(f"Could not attach edge {edge_json['_item']} to {item}, {e}")
                continue
        return item

    def search_last_added(self, type=None, with_prop=None, with_val=None):
        query = {"_limit": 1, "_sortOrder": "Desc"}
        if type is not None:
            query["type"] = type
        if with_prop is not None:
            query[f"{with_prop}=="] = with_val
        return self.search(query)[0]

    def item_from_json(
        self,
        json: dict,
        add_to_local_db: bool = True,
        priority=None,
    ) -> ItemBase:
        priority = Priority(priority) if priority else None

        item_class = get_schema_cls(
            json["type"],
            extra=self.registered_classes,
        )

        new_item = item_class.from_json(json)
        if add_to_local_db:
            try:
                new_item = self.add_to_store(new_item, priority=priority)
            except Exception as e:
                logger.error(f"Could not add {new_item} to local db, {e}")

        if hasattr(new_item, "requires_client_ref") and new_item.requires_client_ref:
            new_item._client = self
        return new_item

    def _item_from_graphql(self, data):
        item = self.item_from_json(data)
        for prop in item.edges:
            if prop in data:
                for edge in data[prop]:
                    edge_item = self._item_from_graphql(edge)
                    item.add_edge(prop, edge_item)
        return item

    def search_graphql(
        self, query: Union[str, GQLQuery], variables: Optional[Dict[str, Any]] = None
    ) -> List[ItemBase]:
        response = self.api.graphql(query, variables)
        data = response["data"]
        result = []
        for d in data:
            item = self._item_from_graphql(d)
            result.append(item)
        return result

    def send_email(self, to, subject="", body=""):
        try:
            self.api.send_email(to, subject, body)
            logger.info(f"succesfully sent email to {to}")
            return True
        except PodError as e:
            logger.error(e)
            return False

    def sync(self, priority: str = Priority.newest):
        priority = Priority(priority) if priority else None
        all_items = list(self.local_db.nodes.values())
        all_ids = set(self.local_db.nodes.keys())

        create_items = []
        update_items = []
        create_edges = []

        # Add items from sync store
        for item in all_items:
            if isinstance(item, File) or isinstance(item, Photo):
                continue
            if item._in_pod:
                update_items.append(item)
            else:
                create_items.append(item)

            # Add all edges where src and tgt exist
            for edge in item._new_edges:
                if edge.target.id in all_ids:
                    create_edges.append(edge)
                else:
                    warnings.warn(
                        f"Could not sync `{edge.name}` for {item}: edge target missing in sync."
                    )

        update_ids = [item.id for item in update_items]
        existing_items = self.search({"ids": update_ids}, add_to_local_db=True, priority=priority)

        return self.bulk_action(
            create_items=create_items,
            update_items=update_items,
            create_edges=create_edges,
            priority=priority,
        )

    def get_dataset(self, name):
        datasets = self.search({"type": "Dataset", "name": name})
        if len(datasets) == 0:
            raise PodError(f"No datasets found with name {name}")
        elif len(datasets) > 1:
            warnings.warn(f"Multiple datasets found with name {name}. Using the newest dataset.")
        return datasets[-1]

    def send_trigger_status(self, item_id, trigger_id, status):
        try:
            self.api.send_trigger_status(item_id, trigger_id, status)
            return True
        except Exception as e:
            logger.error("Failed to send trigger status to the POD")
            return False

    def get_oauth2_access_token(self, platform):
        try:
            return self.api.oauth2get_access_token(platform)["accessToken"]
        except PodError as e:
            logger.error(e)
            return None

    def plugin_status(self, plugins):
        try:
            return self.api.plugin_status(plugins)["plugins"]
        except PodError as e:
            logger.error(e)
            return None

    def get_oauth1_request_token(self, platform, callback_url):
        return self.api.oauth1_request_token(platform, callback_url)

    def get_oauth1_access_token(self, oauth_token, oauth_verifier, oauth_token_secret):
        return self.api.oauth1_access_token(
            oauth_token=oauth_token,
            oauth_verifier=oauth_verifier,
            oauth_token_secret=oauth_token_secret,
        )

    def oauth2_authorize(self, *, platform, code, redirect_uri, pkce_verifier) -> str:
        try:
            return self.api.oauth2authorize(
                code=code,
                platform=platform,
                redirect_uri=redirect_uri,
                pkce_verifier=pkce_verifier,
            )["accessToken"]
        except PodError as e:
            logger.error(e)
            return None

    def get_oauth2_authorization_url(self, platform, scopes, redirect_uri) -> Dict[str, str]:
        try:
            return self.api.oauth2get_authorization_url(platform, scopes, redirect_uri)
        except PodError as e:
            logger.error(e)
            return None



from ...data.basic import HOME_DIR, read_json

PLUGIN_DIR = HOME_DIR / ".pymemri" / "plugins"


def read_username_password(plugin: str):
    credentials_path = PLUGIN_DIR / plugin / "credentials.json"
    credentials = read_json(credentials_path)
    username, password = credentials["username"], credentials["password"]
    return username, password



from pymemri.data.schema import Message
from pymemri.plugin.pluginbase import PluginBase


class ExamplePlugin(PluginBase):
    def __init__(self, **kwargs):
        print("Initializing plugin...")
        super().__init__(**kwargs)

    def run(self):
        print("Started plugin run...")
        message = Message(content="test")
        self.client.create(message)
        print("Run success!")



import io
import json
from hashlib import sha256
from pathlib import Path
from typing import Any, List, Optional, Tuple

import numpy as np
from PIL import ExifTags, Image

from . import _central_schema as cs

DEFAULT_ENCODING = "PNG"


class Photo(cs.Photo):
    data: Optional[bytes] = None

    def show(self):
        raise NotImplementedError()

    @property
    def size(self):
        return self.width, self.height

    @classmethod
    def from_path(cls, path: Path, size: Optional[Tuple[int]] = None):
        pil_image = Image.open(path)
        if size is not None:
            pil_image = pil_image.resize(size)
        return cls.from_PIL(pil_image)

    @classmethod
    def from_np(cls, data: np.array, size: Optional[Tuple[int]] = None):
        pil_image = Image.fromarray(data)
        if size is not None:
            pil_image = pil_image.resize(size)
        return cls.from_PIL(pil_image)

    @classmethod
    def from_bytes(cls, _bytes, size: Optional[Tuple[int]] = None):
        image_stream = io.BytesIO(_bytes)
        pil_image = Image.open(image_stream)
        if size is not None:
            pil_image = pil_image.resize(size)
        return cls.from_PIL(pil_image, _bytes)

    @classmethod
    def from_PIL(cls, image: Image.Image, bytes_: Optional[bytes] = None):
        encoding, mode, shape = cls.infer_PIL_metadata(image)
        w, h, c = shape
        exif_data = cls.extract_exif(image)
        _bytes = bytes_ or cls.PIL_to_bytes(image, encoding)

        res = cls(
            data=_bytes,
            height=h,
            width=w,
            channels=c,
            encoding=encoding,
            mode=mode,
            exifData=exif_data,
            id=cls.create_id(),
        )
        file = cs.File(sha256=sha256(_bytes).hexdigest())
        res.add_edge("file", file)
        file.id = file.create_id()
        return res

    @staticmethod
    def PIL_to_bytes(pil_image, encoding):
        byte_io = io.BytesIO()
        pil_image.save(byte_io, encoding)
        return byte_io.getvalue()

    @staticmethod
    def extract_exif(pil_image: Image.Image) -> Optional[str]:
        # extract exif
        exif_dict = {}
        exif_data = pil_image.getexif()
        for k, v in exif_data.items():
            if k in ExifTags.TAGS:
                exif_dict[ExifTags.TAGS[k]] = str(v)  # may include bytes
        if len(exif_dict.keys()) == 0:
            exif_dict = None
        return json.dumps(exif_dict)

    @staticmethod
    def infer_PIL_metadata(pil_image: Image.Image):
        encoding = pil_image.format or DEFAULT_ENCODING
        mode = pil_image.mode
        size = pil_image.size
        if len(size) == 3:
            w, h, c = size
        elif len(size) == 2:
            w, h = size
            c = 1
        else:
            raise ValueError()
        return encoding, mode, (w, h, c)

    def to_PIL(self) -> Image.Image:
        if self.data is None:
            raise ValueError("Photo object has no data")

        return Image.open(io.BytesIO(self.data))

    def to_np(self) -> np.array:
        pil_img = self.to_PIL()
        return np.asarray(pil_img)


class File(cs.File):
    def __hash__(self):
        return hash(self.sha256)



import json
import os
import time
import traceback
from pathlib import Path

from fastscript import Param, call_parse
from loguru import logger

from pymemri import __version__ as pymemri_version

from ..data.basic import write_json
from ..data.schema import PluginRun
from ..pod.client import DEFAULT_POD_ADDRESS, DEFAULT_POD_KEY_PATH, PodClient
from ..pod.utils import DEFAULT_POD_KEY_PATH, POD_KEYS_FULL_FOLDER, read_pod_key
from .constants import (
    POD_AUTH_JSON_ENV,
    POD_FULL_ADDRESS_ENV,
    POD_OWNER_KEY_ENV,
    POD_PLUGIN_DNS_ENV,
    POD_TARGET_ITEM_ENV,
)
from .oauth_handler import run_twitter_oauth_flow
from .pluginbase import (
    PluginError,
    create_run_expanded,
    parse_config,
    parse_metadata,
    run_plugin_from_run_id,
    write_run_info,
)
from .states import RUN_FAILED

logger.info(f"Pymemri version: {pymemri_version}")


def _parse_env():
    env = os.environ
    logger.info("Reading `run_plugin()` parameters from environment variables")
    try:
        pod_full_address = env.get(POD_FULL_ADDRESS_ENV, DEFAULT_POD_ADDRESS)
        plugin_run_json = json.loads(str(env[POD_TARGET_ITEM_ENV]))
        logger.debug(plugin_run_json)
        plugin_run_id = plugin_run_json["id"]
        owner_key = env.get(POD_OWNER_KEY_ENV)
        pod_auth_json = json.loads(str(env.get(POD_AUTH_JSON_ENV)))
        return pod_full_address, plugin_run_id, pod_auth_json, owner_key
    except KeyError as e:
        raise Exception("Missing parameter: {}".format(e)) from None


@call_parse
def store_keys(
    path: Param("path to store the keys", str) = DEFAULT_POD_KEY_PATH,
    database_key: Param("Database key of the pod", str) = None,
    owner_key: Param("Owner key of the pod", str) = None,
    replace: Param("Replace existing stored keys", str) = True,
):
    if not replace:
        try:
            read_pod_key("database_key")
            read_pod_key("owner_key")
            logger.info("Existing stored keys found, exiting without generating new keys.")
            return
        except ValueError:
            pass

    if database_key is None:
        database_key = PodClient.generate_random_key()
    if owner_key is None:
        owner_key = PodClient.generate_random_key()

    obj = {"owner_key": owner_key, "database_key": database_key}
    Path(path).parent.mkdir(parents=True, exist_ok=True)
    if path.exists():
        timestr = time.strftime("%Y%m%d-%H%M%S")
        path.rename(POD_KEYS_FULL_FOLDER / f"keys-{timestr}.json")
    write_json(obj, path)


@call_parse
def run_plugin(
    pod_full_address: Param("The pod full address", str) = DEFAULT_POD_ADDRESS,
    plugin_run_id: Param("Run id of the plugin to be executed", str) = None,
    database_key: Param("Database key of the pod", str) = None,
    owner_key: Param("Owner key of the pod", str) = None,
    read_args_from_env: Param("Read the args from the environment", bool) = False,
    metadata: Param("metadata file for the PluginRun", str) = None,
    config_file: Param(
        "A plugin configuration, overwrites the configuration of the PluginRun", str
    ) = None,
):
    if read_args_from_env:
        pod_full_address, plugin_run_id, pod_auth_json, owner_key = _parse_env()
        database_key = None
    else:
        if database_key is None:
            database_key = read_pod_key("database_key")
        if owner_key is None:
            owner_key = read_pod_key("owner_key")
        pod_auth_json = None
    if POD_PLUGIN_DNS_ENV in os.environ:
        print(f"Plugin accessible via {os.environ.get(POD_PLUGIN_DNS_ENV)}:8080")

    client = PodClient(
        url=pod_full_address,
        database_key=database_key,
        owner_key=owner_key,
        auth_json=pod_auth_json,
    )
    print(f"pod_full_address={pod_full_address}\nowner_key={owner_key}\n")

    if metadata is not None:
        run = parse_metadata(metadata, remove_container=True)
        create_run_expanded(client, run)
        plugin_run_id = run.id
    else:
        run = client.get(plugin_run_id)
    plugin_config = parse_config(run.config, config_file)

    try:
        run_plugin_from_run_id(plugin_run_id, client, **plugin_config)
    except Exception as e:
        run = client.get(plugin_run_id)
        run.status = RUN_FAILED
        client.update_item(run)
        print(traceback.format_exc(), flush=True)
        raise PluginError("The plugin quit unexpectedly.") from None


@call_parse
def simulate_run_plugin_from_frontend(
    pod_full_address: Param("The pod full address", str) = DEFAULT_POD_ADDRESS,
    database_key: Param("Database key of the pod", str) = None,
    owner_key: Param("Owner key of the pod", str) = None,
    container: Param("Pod container to run frod", str) = None,
    plugin_path: Param("Plugin path", str) = None,
    metadata: Param("metadata file for the PluginRun", str) = None,
    config_file: Param(
        "A plugin configuration, overwrites the configuration of the PluginRun", str
    ) = None,
    account_id: Param("Account id to be used inside the plugin", str) = None,
):
    if database_key is None:
        database_key = read_pod_key("database_key")
    if owner_key is None:
        owner_key = read_pod_key("owner_key")
    params = [pod_full_address, database_key, owner_key]
    if None in params:
        raise ValueError("Missing Pod credentials")

    client = PodClient(url=pod_full_address, database_key=database_key, owner_key=owner_key)
    print(f"pod_full_address={pod_full_address}\nowner_key={owner_key}\n")

    if metadata is not None:
        run = parse_metadata(metadata)
        create_run_expanded(client, run)
    else:
        if container is None:
            container = plugin_path.split(".", 1)[0]
        print(f"Inferred '{container}' as plugin container name")
        plugin_module, plugin_name = plugin_path.rsplit(".", 1)
        run = PluginRun(container, plugin_module, plugin_name)

        if account_id is not None:
            account = client.get(account_id)
            run.add_edge("account", account)
            print(f"Using existing {account}")

        client.create(run)

    print(f"Created pluginrun with id {run.id} on {pod_full_address}")

    plugin_dir = run.containerImage
    write_run_info(plugin_dir, run.id)

    print("*Check the pod log/console for debug output.*")
    return run


@call_parse
def simulate_oauth1_flow(
    pod_full_address: Param("The pod full address", str) = DEFAULT_POD_ADDRESS,
    port: Param("Port to listen on", int) = 3667,
    host: Param("Host to listen on", str) = "localhost",
    callback_url: Param("Callback url", str) = None,
    database_key: Param("Database key of the pod", str) = None,
    owner_key: Param("Owner key of the pod", str) = None,
    metadata: Param("metadata file for the PluginRun", str) = None,
):
    if database_key is None:
        database_key = read_pod_key("database_key")
    if owner_key is None:
        owner_key = read_pod_key("owner_key")
    if metadata is None:
        raise ValueError("Missing metadata file")
    else:
        run = parse_metadata(metadata)
    params = [pod_full_address, database_key, owner_key]
    if None in params:
        raise ValueError("Missing Pod credentials")

    print(f"pod_full_address={pod_full_address}\nowner_key={owner_key}\n")
    client = PodClient(url=pod_full_address, database_key=database_key, owner_key=owner_key)

    if run.pluginName == "TwitterPlugin":
        run_twitter_oauth_flow(client=client, host=host, port=port, callback_url=callback_url)
    else:
        raise ValueError("Unsupported plugin")



import fnmatch
import json
import logging
import os
import platform
import socket
import urllib
from collections import deque
from hashlib import sha256
from typing import Any, Deque, Dict, Generator, List, Optional, Union

import requests
from loguru import logger
from urllib3.connection import HTTPConnection

from .graphql_utils import GQLQuery

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("urllib").setLevel(logging.WARNING)
logging.basicConfig(
    format="%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d:%H:%M:%S",
    level=logging.DEBUG,
)

DEFAULT_POD_ADDRESS = os.environ.get("POD_ADDRESS") or "http://localhost:3030"
POD_VERSION = "v4"
POD_ALLOWED_ORIGINS = os.environ.get("POD_ALLOWED_ORIGINS", "*").split(",")


class PodError(Exception):
    def __init__(self, status=None, message=None, **kwargs) -> None:
        super().__init__(status, message, **kwargs)
        self.status = status
        self.message = message

    def __str__(self) -> str:
        return " ".join([str(a) for a in self.args if a])


def get_platform_specific_keepalive_option():
    if platform.system() == "Darwin":
        return (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
    else:
        return (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 60)


class PodAPI:
    def __init__(
        self,
        database_key: str,
        owner_key: str,
        url: str = DEFAULT_POD_ADDRESS,
        version: str = POD_VERSION,
        auth_json: dict = None,
        verbose: bool = True,
    ) -> None:
        self._check_origin(url)

        self.verbose = verbose
        self.database_key = database_key
        self.owner_key = owner_key
        self.version = version
        self._url = url

        self.base_url = f"{url}/{version}/{self.owner_key}"
        self.base_url_v5 = f"{url}/v5/{self.owner_key}"
        self.auth_json = self._create_auth(auth_json)

        # Class uses requests Session to communicate with the POD.
        # That creates a pool of TCP connections, over which HTTP requests
        # with Keep-Alive header are sent. While idle, there is no traffic
        # on those connections, that gets detected by the network balancer
        # of the AWS infrastructure and makes those connection dropped,
        # resulting with "connection reset by peer", or similar errors
        # visible on the client side.
        # Enabling TCP keep-alive packets resets the network balancer idle timer
        # resulting in persistent connection as it was originally intended.
        HTTPConnection.default_socket_options = HTTPConnection.default_socket_options + [
            # Enable TCP keepalive packet transmission
            (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
            # Start sending after 60 sec of idleness
            get_platform_specific_keepalive_option(),
            # Send keep-alive at 60 sec interval
            (socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 60),
            # Close connection after 5 failed keep-alive pings (5 minutes)
            (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5),
        ]

        self.session = requests.Session()

    def _check_origin(self, url):
        for origin in POD_ALLOWED_ORIGINS:
            if fnmatch.fnmatch(url, origin):
                return

        raise PodError(
            403,
            f"Trying to create POD API client with callback url {url} that is outside allowed origins {POD_ALLOWED_ORIGINS}",
        )

    def _create_auth(self, auth_json: dict = None) -> dict:
        if auth_json is not None:
            return {"type": "PluginAuth", **auth_json}
        else:
            return {"type": "ClientAuth", "databaseKey": self.database_key}

    def create_account(self):
        response = self.session.post(
            f"{self._url}/{self.version}/account",
            json={"ownerKey": self.owner_key, "databaseKey": self.database_key},
        )
        # if response.status_code != 200:
        #     raise PodError(response.status_code, response.text)

    def test_connection(self) -> bool:
        try:
            res = self.session.get(self._url)
            if self.verbose:
                logger.info("Successfully connected to pod")
            return True
        except requests.exceptions.RequestException as e:
            logger.error("Could no connect to backend")
            return False

    @property
    def pod_version(self) -> dict:
        response = self.session.get(f"{self._url}/version")
        if response.status_code != 200:
            raise PodError(response.status_code, response.text)
        return response.json()

    def post(self, endpoint: str, payload: Any) -> Any:
        body = {"auth": self.auth_json, "payload": payload}
        response = self.session.post(f"{self.base_url}/{endpoint}", json=body)
        if response.status_code != 200:
            raise PodError(response.status_code, response.text)
        return response

    def post_v5(self, endpoint: str, payload: Any) -> Any:
        body = {"auth": self.auth_json, "payload": payload}
        response = self.session.post(f"{self.base_url_v5}/{endpoint}", json=body)
        if response.status_code != 200:
            raise PodError(response.status_code, response.text)
        return response

    def create_schema(self, item: dict) -> str:
        return self.post_v5("schema", item).json()

    def get_item(self, uid: str) -> dict:
        return self.post("get_item", uid).json()

    def create_item(self, item: dict) -> str:
        return self.post("create_item", item).json()

    def update_item(self, item: dict) -> list:
        return self.post("update_item", item).json()

    def get_edges(
        self, uid: str, direction: str = "Outgoing", expand_items: bool = True
    ) -> List[dict]:
        payload = {"item": uid, "direction": direction, "expandItems": expand_items}
        return self.post("get_edges", payload).json()

    def create_edge(self, edge: dict) -> str:
        return self.post("create_edge", edge).json()

    def delete_item(self, uid) -> list:
        return self.post("delete_item", uid).json()

    def delete_edge(self, edge: dict) -> str:
        return self.post("delete_edge_by_source_target", edge).json()

    def delete_file(self, id: str) -> str:
        return self.post("delete_file", {"id": id}).json()

    def search(self, query: dict) -> List[dict]:
        return self.post("search", query).json()

    def search_paginate(self, query: dict, limit: int = 32, even_page_size=True) -> Generator:
        """
        The Pod returns uneven page sizes when paginating, which can be an issue for some applications.
        `search_paginate` wraps the pagination, and always returns pages of size `limit` by storing overflow items
        in a queue.
        """
        paginator = self._paginate(query, limit)

        if not even_page_size:
            yield from paginator

        remaining: Deque[Dict[str, Any]] = deque()
        while True:
            if len(remaining) >= limit:
                yield [remaining.popleft() for _ in range(limit)]
            if len(remaining) < limit:
                try:
                    remaining.extend(next(paginator))
                except StopIteration:
                    break

        while len(remaining):
            yield [remaining.popleft() for _ in range(min(limit, len(remaining)))]

    def _paginate(self, query: dict, limit: int = 32) -> Generator:
        if (
            "_limit" in query
            or "dateServerModified" in query
            or "dateServerModified>=" in query
            or "dateServerModified<" in query
        ):
            raise ValueError("Cannot paginate query that contains a date or limit.")
        if "_sortOrder" in query:
            raise NotImplementedError("Only 'Asc' order is supported.")

        query = {**query, "_limit": limit}
        response = self.search(query)
        if not len(response):
            return
            yield

        offset = 0
        query = {**query, "_limit": limit}

        while True:
            query["_offset"] = offset

            response = self.search(query)
            if not len(response):
                break

            offset += limit
            yield response

    def bulk(
        self,
        create_items: List[dict] = None,
        update_items: List[dict] = None,
        create_edges: List[dict] = None,
        delete_items: List[str] = None,
        search: List[dict] = None,
    ) -> Dict[str, Any]:
        payload = {
            "createItems": create_items,
            "updateItems": update_items,
            "createEdges": create_edges,
            "deleteItems": delete_items,
            "search": search,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        return self.post("bulk", payload).json()

    def graphql(
        self, query: Union[str, GQLQuery], variables: Optional[Dict[str, Any]] = None
    ) -> List[dict]:
        if isinstance(query, str):
            query = GQLQuery(query)
        query.format(variables)
        return self.post("graphql", query.data).json()

    def oauth(self, service: str, callback_url: str):
        payload = {"service": service, "callbackUrl": callback_url}
        return self.post("oauth", payload).json()

    def upload_file(self, file: bytes) -> Any:
        if self.auth_json.get("type") == "PluginAuth":
            # alternative file upload for plugins, with different authentication
            return self.upload_file_b(file)

        sha = sha256(file).hexdigest()
        result = self.session.post(
            f"{self.base_url}/upload_file/{self.database_key}/{sha}", data=file
        )
        if result.status_code != 200:
            raise PodError(result.status_code, result.text)

        return result

    def upload_file_b(self, file: bytes) -> Any:
        sha = sha256(file).hexdigest()
        auth = urllib.parse.quote(json.dumps(self.auth_json))
        result = self.session.post(f"{self.base_url}/upload_file_b/{auth}/{sha}", data=file)
        if result.status_code != 200:
            raise PodError(result.status_code, result.text)

        return result

    def get_file(self, sha: str) -> bytes:
        return self.post("get_file", {"sha256": sha}).content

    def send_email(self, to: str, subject: str = "", body: str = "") -> Any:
        payload = {"to": to, "subject": subject, "body": body}
        return self.post("send_email", payload)

    def send_trigger_status(self, item_ids: List[str], trigger_id: str, status: str) -> Any:
        payload = {"item_ids": item_ids, "trigger_id": trigger_id, "status": status}
        return self.post("trigger/status", payload)

    def oauth2get_access_token(
        self,
        platform: str,
    ) -> Dict:
        return self.post(
            "oauth2/access_token",
            {"platform": platform},
        ).json()

    def oauth2authorize(
        self, *, platform: str, code: str, redirect_uri: str, pkce_verifier: str
    ) -> Dict[str, str]:
        return self.post(
            "oauth2/authorize",
            {
                "platform": platform,
                "authCode": code,
                "redirectUri": redirect_uri,
                "pkceVerifier": pkce_verifier,
            },
        ).json()

    def oauth2get_authorization_url(
        self, platform: str, scopes: str, redirect_uri: str
    ) -> Dict[str, str]:
        return self.post(
            "oauth2/auth_url",
            {
                "platform": platform,
                "scopes": scopes,
                "redirectUri": redirect_uri,
            },
        ).json()

    def oauth1_request_token(self, platform: str, callback_url: str) -> Any:
        return self.post(
            "oauth1_request_token", {"service": platform, "callbackUrl": callback_url}
        ).json()

    def oauth1_access_token(self, *, oauth_token, oauth_token_secret, oauth_verifier) -> Any:
        return self.post(
            "oauth1_access_token",
            {
                "oauthVerifier": oauth_verifier,
                "oauthToken": oauth_token,
                "oauthTokenSecret": oauth_token_secret,
            },
        ).json()

    def plugin_status(self, plugins: List[str]) -> Any:
        return self.post("plugin/status", {"plugins": plugins}).json()

    def plugin_api(self, plugin_id: str) -> Any:
        return self.post("plugin/api", {"id": plugin_id}).json()

    def plugin_api_call(
        self,
        plugin_id: str,
        method: str,
        endpoint: str,
        query: Dict[str, Any] = None,
        jsonBody: Dict[str, Any] = None,
    ) -> Any:
        payload = {
            "id": plugin_id,
            "method": method,
            "endpoint": endpoint,
            "query": query,
            "jsonBody": jsonBody,
        }

        return self.post("plugin/api/call", payload).json()
